#ifndef _FL_COLUMNMANAGER_
#define _FL_COLUMNMANAGER_

#include <FL/Fl.H>
#include <FL/Fl_Widget.H>
#include <FL/Fl_Browser.H>
#include <FL/fl_draw.H>
#include <stdlib.h>
#include <string>
#include <vector>
#include <iostream>

//! An Fl_ColumnManager controls the row sorting of an Fl_Browser.
/*! The class is designed to manage the row sorting of
  the text in an Fl_Browser. Depressing any column button
  will cause the rows to be sorted by the text in that column.
  It uses the column_widths() of the associated browser to draw itself.
  These widths can be changed interactively.
*/
class Fl_ColumnManager : public Fl_Widget {
public:
  //! Standard fltk constructor.
  Fl_ColumnManager(int x, int y, int w, int h, const char *label=NULL);
  //! Process click and drag events.
  int handle(int event);
  //! Force sorting and redraw of controller.
  void update();
  //! Identify the associated browser and establish a link with the it.
  void set_browser(Fl_Browser* b);
  //! set the specified column header
  void set_header(int c, const std::string& h);
  //! Respond to fltk draw requests.
  void draw();
protected:
  //! Draw control button for the specified column, it's box must be passed in.
  void draw_control(int c, int X, int Y, int W, int H);
  //! Process a click for a specified control button, it's box must be passed in.
  void click_control(int c, int X, int Y, int W, int H);
  //! Process a drag for a specified divider, the lefthand control button's box must be passed in.
  void drag_control(int c, int X, int Y, int W, int H);
  //! Replacement callback for the browser's hscrollbar
  static void scrollbar_callback(Fl_Widget* s, void* cm);
  //! Common utility for processing each column.
  void for_each_column(void (Fl_ColumnManager::*action)(int col, int,int,int,int));
  // member data
  Fl_Browser* B;     //!< the associated browser
  int mySortColumn;  //!< the column to sort by
  int myDrag;        //!< the separator being dragged
  std::vector<STL string> headers; //!< column labels
};

//! count max columns
/*! uses the browser's column_char to identify columns */
extern int max_columns(const Fl_Browser* B);

//! simple function to compare two strings.
extern bool strictly_less(const STL string& A, const STL string& B);

//! sorts the rows of an Fl_Browser using the lex function on the specified column text
extern void sort_by_column(int c, Fl_Browser* B, bool (*lex)(const STL string&,const STL string&)=&strictly_less);

//! export to the stream with html table or text formatting
/*! if the col_separator is NULL then html is output,
  otherwise it is used to separate columns in a text file.
  If headers is not NULL then it will be used to label the columns.
*/
extern void export_table(const Fl_Browser* B, const char *headers[], std::ostream& ostrm, const char* delim=NULL);
 
#endif /* _FL_COLUMNMANAGER_ */
