#include "FL/Fl_ColumnManager.H"

/*! Defers associating the controlled browser.
 Sets the sorting column to none (-1). */
Fl_ColumnManager::Fl_ColumnManager(int x, int y, int w, int h, const char *label)
  : Fl_Widget(x, y, w, h, label),
    B(NULL),            
    mySortColumn(-1),
    myDrag(-1),
    headers()
{
}

/*! FL_PUSH checks for a click on each column control.
    FL_DRAG tries to drag the column divider, if close enough (+/-3 pixels).
*/
int Fl_ColumnManager::handle(int event)
{
  if(B!=NULL) switch(event) {
  case FL_PUSH:
    for_each_column(&Fl_ColumnManager::click_control);
    myDrag = -1; // clear current dragging (possible start of new dragging)
    return 1;
  case FL_DRAG:
    for_each_column(&Fl_ColumnManager::drag_control);
    return 1;
  }
  return 0;
}

/*! Performs the same task as the fltk 1.1.4 callback does,
  but also forces the control manager to redraw. */
void Fl_ColumnManager::scrollbar_callback(Fl_Widget* s, void* cm)
{
  ((Fl_Browser_*)(s->parent()))->hposition(int(((Fl_Scrollbar*)s)->value()));
  ((Fl_ColumnManager*)cm)->damage(FL_DAMAGE_ALL);
}


void Fl_ColumnManager::set_header(int c, const std::string& h)
{
  if(c+1>(int)headers.size()) headers.resize(c+1);
  headers[c] = h;
}

/*! In addition to identifying the associated browser,
  the callback for the browser's hscrollbar is changed.
  This also sets the hscrollbar's user_data to point to this column manager.*/
void Fl_ColumnManager::set_browser(Fl_Browser* b)
{
  B=b;
  damage(FL_DAMAGE_ALL);
  if(B!=NULL) B->hscrollbar.callback(scrollbar_callback,this);
}

/*! Could be modified to draw a label in the control button. */
void Fl_ColumnManager::draw_control(int c, int X, int Y, int W, int H)
{
  fl_draw_box(c==mySortColumn?FL_DOWN_BOX:FL_UP_BOX,X,Y,W,H,color());
  if(c<(int)headers.size()) {
    fl_push_clip(X,Y,W,H);
    fl_color(FL_BLACK);
    fl_draw(headers[c].c_str(),X+2,Y+1,W-4,H-2,(Fl_Align)(FL_ALIGN_LEFT|FL_ALIGN_CLIP),NULL,false); // false ==> no special symbols marked by '@'
    fl_pop_clip();
  }
}

/*! Excludes 3 pixels at the ends of the control button for resizing them. */
void Fl_ColumnManager::click_control(int c, int X, int Y, int W, int H)
{
  if(Fl::event_x()>X+3 && Fl::event_x()-X<W-6) {
    mySortColumn =  mySortColumn == c? -1 : c;    // toggle off if already selected
    update();
  }
}

/*! Resizing via dragging requires the mouse be within 3 pixels of the
  control button dividers. */
void Fl_ColumnManager::drag_control(int c, int X, int Y, int W, int H)
{
  X = Fl::event_x() - (X+W);
  if(c==myDrag || (-3<=X && X<=3)) {
    myDrag=c; // remember which column is being dragged
    const int *w = B->column_widths();
    if(w[c]>0 && w[c]+X>2) { // if zero then last column which is not adjustable
      ((int*)w)[c] += X;
      damage(FL_DAMAGE_ALL);
      B->damage(FL_DAMAGE_ALL);
    }
  }
}

/*! This should be called whenever the sorting needs to be done
  or the column widths have been changed directly through the browser. */
void Fl_ColumnManager::update()
{
  damage(FL_DAMAGE_ALL);
  sort_by_column(mySortColumn,B);
}

/*! Uses the common utility for processing each column. */
void Fl_ColumnManager::draw()
{
  if (damage()) for_each_column(&Fl_ColumnManager::draw_control);
}


/*! Each column's box is calculated and then the action called for it. */
void Fl_ColumnManager::for_each_column(void (Fl_ColumnManager::*action)(int col, int,int,int,int))
{
  int X = x(), Y = y(), W = w();
  if(B!=NULL) {
    int S = B->hposition();
    const int* C = B->column_widths();
    for(int c=0;;c++) {
      int wi = *C>0? *C-S : W-X;
      if(wi>0) {
	(this->*action)(c,X,Y,wi,h());
	X += wi;
	S = 0;
      } else S = -wi;
      if(*C++ == 0) break;
    }
  } else (this->*action)(0,X,Y,W,h());
}

  


//! Returns a copy of the text in the specified column of the given concatenated text.
STL string column_text(int c, const char* ln, const char COL_MARKER='\t')
{
  if(ln==NULL) return "";
  while(c>0 && *ln!=0) if(*ln++==COL_MARKER) c--;
  if(ln==NULL) return "";
  int n=0;
  while(ln[n]!=0) if(ln[n++]==COL_MARKER) break;
  return STL string(ln,n);
}

//! Returns a copy of the text in the specified row and column of an Fl_Browser.
STL string column_text(int r, int c, const Fl_Browser* B)
{
  return column_text(c,B->text(r),B->column_char());
}

//! count max columns
int max_columns(const Fl_Browser* B)
{
  int n=0; char d = B->column_char();
  for(int r=0;r<B->size();r++) {
    int c=0;
    const char* ln = B->text(r);
    if(ln==NULL) continue;
    while(*ln!=0) if(*ln++==d) c++;
    if(c>n) n=c;
  }
  return n;
}

//! Swaps two rows maintaining their selection states.
/*! This routine would be difficult to use when sorting rows when
  desiring to preserve order of "equal" rows. */
// WARNING Fl_Browser::selected(int) is not documented (fltk 1.1.4), only some derived classes document it
void swap_rows(int r1,int r2, Fl_Browser* B)
{
  STL string  t1 = B->text(r1),      t2 = B->text(r2); // must local make copy
        void *d1 = B->data(r1),     *d2 = B->data(r2);
	int   s1 = B->selected(r1),  s2 = B->selected(r2);
  B->text(r1,t2.c_str()); B->data(r1,d2); B->select(r1,s2);
  B->text(r2,t1.c_str()); B->data(r2,d1); B->select(r2,s1);
}

//! Removes the row 'fr' and inserts it before 'to' maintaining its selection state.
/*! After the call the row originally at 'fr' is at 'to'.
  Only the relative ordering of the moved row is changed.
  Thus, it can be used to preserve order of "equal" rows when sorting. */
// WARNING Fl_Browser::selected(int) is not documented (fltk 1.1.4), only some derived classes document it
void move_row(int fr,int to, Fl_Browser* B)
{
  STL string  t = B->text(fr); // must local make copy
  void *d = B->data(fr);
  int   s = B->selected(fr);
  B->remove(fr);
  B->insert(to,t.c_str(),d);
  B->select(to,s);
}


bool strictly_less(const STL string& A, const STL string& B)
{
  return A<B;
}


/*! The sort is not terribly efficient but perserves the existing order of equal rows.
  Also, the column text is not cached. */
void sort_by_column(int c, Fl_Browser* B, bool (*lex)(const STL string&,const STL string&))
{
  if(c<0) return; // no sorting requested
  for(int a=1,a0,N=B->size();a<=N;a++) {
    STL string ta = column_text(a,c,B); a0=a;
    // find the smallest of the remaining ones
    for(int b=a+1;b<=N;b++) {
      STL string tb = column_text(b,c,B);
      if(lex(tb,ta)) {ta=tb; a0=b;}
    }
    if(a0!=a) move_row(a0,a,B); // swap would destroy order
  }
}


void export_table(const Fl_Browser* B, const char *headers[], std::ostream& ostrm, const char* delim)
{
  static std::string html_head =
    "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\n"
    "<html>\n"
    "<head>\n"
    "<title>Datants - Exported Table</title>\n"
    "<meta http-equiv=\"content-type  content=\"text/html; charset=ISO-8859-1\">\n"
    "</head>\n"
    "<body>\n"
    "<table cellpadding=2 cellspacing=2 border=1>\n";
  static std::string html_tail = "\n</body>\n</html>\n";
  bool html = delim==NULL;
  int ncol = max_columns(B);
  if(html) ostrm<<html_head;
  if(headers!=NULL) {
    for(int c=0;c<ncol;c++) {
      if(html) {
	ostrm<<"  <td style=\"background-color:rgb(255,255,200);";
	//	if(delim!=0) ostrm<<" width:"<<widths[c]<<";";
	ostrm<<"\">"<<headers[c]<<"</td>";
      } else {
	if(c!=0) ostrm<<delim;
	ostrm<<headers[c];
      }
      ostrm << "\n";
    }
  }
  for(int r=0;r<B->size();r++) {
    if(html) ostrm<<"<tr>\n";
    for(int c=0;c<ncol;c++) {
      std::string v = column_text(r,c,B);
      if(html) ostrm<<"  <td>"<<v<<"</td>";
      else {
	if(c!=0) ostrm<<delim;
	ostrm<<v;
      }
    }
    if(html) ostrm<<"</tr>";
    ostrm<<"\n";
  }
  if(html) ostrm<<html_tail;
}
